/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "canvaslayer.h"


ICanvasLayer * g_fac_SPAWN_CANVASLAYER (long width, long height, IOpenGLHelper * helper) {
    return new CanvasLayer (width, height, helper);
}

CanvasLayer::CanvasLayer (long width, long height, IOpenGLHelper * helper) {
    m_Id        = ISpawnedObject::allocateId ();
    m_Width     = width;
    m_Height    = height;
    m_LayerName = tr ("Unnamed");
    m_Opacity   = 1.0f;
    m_Helper    = helper;
    m_Dpi       = 100;
    if (m_Helper) {
        m_Helper->addRef ();
    }

    _register (this);

    m_HasBackground = FALSE;
    m_Background[0] = 0.0f;
    m_Background[1] = 0.0f;
    m_Background[2] = 0.0f;
    m_Background[3] = 0.0f;
}

CanvasLayer::~CanvasLayer (void) {
  //clear ();

    _unregister (this);

    // Release all channels.
    foreach (IChannel * channel, m_Channels) {
        channel->release ();
    }

    SAFE_RELEASE (m_Helper);
}

void CanvasLayer::resort (void) {
    // Selection sort.
    for (int i = 0; i < m_Channels.size (); i++) {
        int swapidx = i;
        IChannel * channel_0 = m_Channels.at (i);
        for (int j = i + 1; j < m_Channels.size (); j++) {
            IChannel * channel_1 = m_Channels.at (j);
            if (channel_1->stackPosition () < channel_0->stackPosition ()) {
                channel_0 = channel_1;
                swapidx = j;
            }
        }

        m_Channels.swap (i, swapidx);
    }
}

bool CanvasLayer::hasChannel (FOURCC uid) {
    foreach (IChannel * channel, m_Channels) {
        if (uid == channel->uid ())
            return TRUE;
    }

    return FALSE;
}

void CanvasLayer::setDpi (int dpi) {
    m_Dpi = dpi;
    foreach (IChannel * channel, m_Channels) {
        channel->setLayerDpi (dpi);
    }
}

/// Touch all channels with this operation.
void CanvasLayer::touch (IInstrumentLockEx * operation, int x, int y) {
    foreach (IChannel * channel, m_Channels) {
        sTouch t;
        t.x = x;
        t.y = y;
        t.feedback = TRUE;
        channel->touch (operation, &t, 1);
    }
}

void CanvasLayer::clear (void) {
    foreach (IChannel * channel, m_Channels) {
        channel->clear ();
    }
}

void CanvasLayer::resize (long width, long height) {
    m_Width  = width;
    m_Height = height;

    foreach (IChannel * channel, m_Channels) {
        channel->resize (width, height);
    }
}

bool CanvasLayer::changesCheck (void) const {
    foreach (IChannel * channel, m_Channels) {
        if (channel->changesCheck ())
            return TRUE;
    }

    return FALSE;
}

void CanvasLayer::changesCommit (void) {
    foreach (IChannel * channel, m_Channels) {
        channel->changesCommit ();
    }
}

unsigned long CanvasLayer::requiredBufferSize (void) const {
    const long layerMetaSz   = sizeof(qint32) * 2 + 256;
    const long channelMetaSz = sizeof(qint32) * 2;

    long sum = layerMetaSz + channelMetaSz * m_Channels.size ();
    foreach (IChannel * channel, m_Channels) {
        sum += channel->requiredBufferSize ();
    }

    return sum;
}

unsigned long CanvasLayer::saveData (QDataStream * stream, unsigned long capacity) {
    if (capacity < requiredBufferSize ())
        return 0L;

    unsigned long written = 0L;

    // Write the properties.
    stream->operator << ((qint32) m_Width);
    stream->operator << ((qint32) m_Height);
    stream->operator << ((qint32) m_Channels.size () - 1);
    stream->operator << (m_Opacity);

    const char * name = m_LayerName.toUtf8 ().data ();

    // And name.
    for (int wr = 0; wr < 256; wr += stream->writeRawData (name + wr, 256 - wr)) ;
    written += sizeof(qint32) * 3 + sizeof(float) + 256;

    // Write the channels data.
    foreach (IChannel * channel, m_Channels) {
        if (capacity > written)
            written += channel->saveData (stream, capacity - written);
    }

    return written;
}

unsigned long CanvasLayer::loadData (QDataStream * stream, unsigned long capacity) {
    clear ();

    // Load everything but dimentions.
    unsigned long read = 0L;
    qint32 channels = 0;

    // Read opacity.
    stream->operator >> (channels);
    stream->operator >> (m_Opacity);
    read += sizeof(float) + sizeof(qint32);

    // Sanity check.
    if (channels > 3)
        return 0L;

    char name[256];

    // And name.
    for (int r = 0; r < 256; r += stream->readRawData (name + r, 256 - r)) ;
    m_LayerName = QString (name);
    read += 256;

    // Read the channels.
    for (int i = 0; i < channels; i++) {
        // Read the metadata.
        qint32 width  = 0;
        qint32 height = 0;
        quint32 uid   = FOURCC_NULL;
        stream->operator >> (width);
        stream->operator >> (height);
        stream->operator >> (uid);
        read += 3 * sizeof(qint32);

        // Sanity check.
        if (width  > 16 * 1024 || width  < 16 ||
            height > 16 * 1024 || height < 16)
            return 0L;
        if (uid != CHANNELUID_Meta0 &&
            uid != CHANNELUID_Optical)
            return 0L;

        // Create a channel.
        IChannel * channel = g_fac_SPAWN_CHANNEL (uid, width, height, m_Helper);

        // Read the channel data.
        if (capacity > read) {
            read += channel->loadData (stream, capacity - read);
        }
        append (channel);
        channel->release ();
    }

    return read;
}
