/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef CANVASLAYER_H
#define CANVASLAYER_H

#include <QString>
#include <QList>
#include <QObject>

#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

#include "../../../glew/include/GL/glew.h"
#include "iCanvasLayer.h"


#define MAX_LAYER_CHANNELS      16

/**
 * A single layer of a canvas.
 * Actually, it is just a wrapper for an
 * independent stack of channels.
 */
class CanvasLayer : public QObject,
                    public CRefCounter,
                    public ICanvasLayer {
    Q_OBJECT

protected:
    /**
     * Channels list.
     */
    QList<IChannel*> m_Channels;

    QString m_LayerName;
    float   m_Opacity;
    float   m_Background[4];
    bool    m_HasBackground;
    long    m_Width;
    long    m_Height;
    int     m_Id;
    int     m_Dpi;

    IOpenGLHelper * m_Helper;

     CanvasLayer (long width, long height, IOpenGLHelper * helper);
    ~CanvasLayer (void);

    void resort (void);

    virtual void suicide (void) {
        delete this;
    }

friend ICanvasLayer * g_fac_SPAWN_CANVASLAYER (long width, long height, IOpenGLHelper * helper);
public:

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::CanvasLayer";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

    /************ OPERATIONS **************/

    /**
     * Apply an instrument at given position.
     */
    virtual void touch (IInstrumentLockEx * operation, int x, int y);

    /**
     * Clear all channels in a layer.
     */
    virtual void clear (void);

    /**
     * Resize the layer.
     */
    virtual void resize (long width, long height);

    /**
     * Get the layer width.
     */
    virtual long width (void) const {
        return m_Width;
    }

    /**
     * Get the layer height.
     */
    virtual long height (void) const {
        return m_Height;
    }

    /********** GET\SET METHODS ***********/

    virtual float opacity (void) const {
        return m_Opacity;
    }

    virtual QString name (void) const {
        return m_LayerName;
    }

    virtual void setName (QString name) {
        m_LayerName = name;
    }

    virtual void setOpacity (float opacity) {
        m_Opacity = opacity;
    }

    /**
     * Check if the layer has some
     * background color.
     */
    virtual bool hasBackground (float * rgba) const {
        if (m_HasBackground && rgba) {
            rgba[0] = m_Background[0];
            rgba[1] = m_Background[1];
            rgba[2] = m_Background[2];
            rgba[3] = m_Background[3];
        }

        return m_HasBackground;
    }

    /**
     * Check if the layer
     * contains the given channel.
     */
    virtual bool hasChannel (FOURCC uid);

    /**
     * Set the layer dpi.
     */
    virtual void setDpi (int dpi);

    /**
     * Get the layer dpi.
     */
    virtual int dpi (void) const {
        return m_Dpi;
    }

    /************** QLIST WRAPPER ***************/

    virtual void append (IChannel * channel) {
        if (m_Channels.size () < MAX_LAYER_CHANNELS) {
            channel->addRef ();
            channel->setLayerDpi (m_Dpi);
            m_Channels.append (channel);
        }

        resort ();
    }

    virtual void prepend (IChannel * channel) {
        if (m_Channels.size () < MAX_LAYER_CHANNELS) {
            channel->addRef ();
            channel->setLayerDpi (m_Dpi);
            m_Channels.prepend (channel);
        }

        resort ();
    }

    virtual void removeAt (int index) {
        if (index >= 0 && index < m_Channels.size ()) {
            m_Channels.at (index)->release ();
            m_Channels.removeAt (index);
        }
    }

    virtual IChannel * at (int index) {
        if (index >= 0 && index < m_Channels.size ())
            return m_Channels.at (index);
        else
            return NULL;
    }

    virtual int size (void) {
        return m_Channels.size ();
    }

    /********* IDATACONTAINER IMPLANTS ***********/

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const;

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void);

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const;

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity);

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity);
};

#endif // CANVASLAYER_H
