/***
 *  Copyright (C) 2009 by Valeriy Fedotov
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "curveedit.h"
#include "commmath.h"

#include <QResizeEvent>
#include <QPainter>
#include <QPointF>
#include <math.h>


CurveEdit::CurveEdit (QWidget *parent, ISpline * source)
    : QWidget(parent)
{
    m_source = NULL;

    resize (150,100);
    setSource (source);
}

/// [castedflake]
/// We need a possibility to change the underlying
/// spline without recreation of widget itself.
void CurveEdit::setSource (ISpline * source) {
    m_points.clear ();

    m_source = source;
    if (m_source) {
        // Init the points by stepping through the spline.
        const float step = 0.2f;
        for (int i = 0; i < 7; i++) {
            m_points.append (QPointF (i * step,
                                      MATH::CLAMP_01 (m_source->f (i * step))
                                     )
                            );
        }

        // Now make the underlying spline consistent
        // with our points set.
        m_source->rebuild (m_points);
    }

    update ();
}

void CurveEdit::paintEvent ( QPaintEvent * )
{
    if (!m_source)
        return;

    int h = height();
    int w = width();
    qreal radius = 5;
    QPainter p(this);
    p.setRenderHint(QPainter::Antialiasing, true);
    p.translate(0, height());
    p.scale(1, -1);

    // Spline drawing.
    p.setPen(Qt::darkGray);
    QPainterPath path(QPointF(0, 0));
    for(double x = 0; x < w; x += 4.0){
        path.lineTo(x, m_source->f(x/w) * h);
    }
    path.lineTo(width(), height());
    p.drawPath(path);

    // Circles drawing.
    p.setPen(Qt::darkCyan);
    for(int i = 1; i < m_points.size() - 1; i++){
        p.drawEllipse(QPointF(m_points[i].x()*width(), m_points[i].y()*height()), radius, radius);
    }

    // Frame drawing.
    p.setPen(Qt::lightGray);
    p.drawRect(0,0, width(), height());
}

void CurveEdit::mousePressEvent ( QMouseEvent * event )
{
    double x = event->posF().x()/width();
    double y = (height() - event->posF().y())/height();
    qreal radius = 5./height();
    int i;
    for(i = 1; i < m_points.size(); i++){
        QPointF p = m_points[i];
        if( sqrt((p.x() - x) * (p.x() - x) + (p.y() - y) * (p.y() - y)) < radius)
            break;
    }
    if(i != m_points.size()){
        m_pressed = true;
        m_movingPointIndex = i;
    }
}

void CurveEdit::mouseReleaseEvent ( QMouseEvent * )
{
    m_pressed = false;
}

void CurveEdit::mouseMoveEvent ( QMouseEvent * event )
{
    if( ! m_pressed )
        return;

    QPointF pos = event->posF();
    pos.setY(height() - pos.y());
    m_points[m_movingPointIndex].setX(pos.x()/width());
    m_points[m_movingPointIndex].setY(pos.y()/height());
    if (m_source)
        m_source->rebuild (m_points);

    update();
}

void CurveEdit::setPoints(QList <QPointF> pointList)
{
    m_points = pointList;
    if (m_source)
        m_source->rebuild (m_points);

    update();
}

QList <QPointF> CurveEdit::points()
{
    return m_points;
}

CurveEdit::~CurveEdit()
{
}
