#include "paperindicator.h"
#include <QPainter>

PaperIndicator::PaperIndicator (QWidget * parent) : QObject (parent) {
    m_Connector = NULL;
    m_LastError = NULL;
    m_Logger    = NULL;

    m_LoggingEnabled  = FALSE;

    m_Id = ISpawnedObject::allocateId ();

    combo = new QComboBox();
    if (combo) {
        combo->setIconSize(QSize(64, 64));
        combo->view()->setIconSize(QSize(32,32));
        static_cast<QListView*>(combo->view())->setViewMode(QListView::IconMode);
        static_cast<QListView*>(combo->view())->setMovement(QListView::Static);
        QObject::connect(combo, SIGNAL(currentIndexChanged(int)), this, SLOT(comboChanged(int)));
    }
}

PaperIndicator::~PaperIndicator (void) {
    clear ();
    if (combo) {
        combo->disconnect();
        delete(combo);
        combo = NULL;
    }

    SAFE_RELEASE (m_Connector);
}

void PaperIndicator::clear (void) {
    if (combo)
        combo->clear ();

    m_TexturesPack.clear ();
}

void PaperIndicator::setPaper (sPaper * dscr) {
    m_LastError = NULL;

    // FIXME I'd better use the color texture here if the item wasn't too small.
    IArray * color = dscr->height;
    const int sz = color->width () * color->height ();
    if (!sz)
        return;

    // Convert the texture to 8-bit intergers.
    BYTE * paperTile_datau = new BYTE[sz * 4 + 32];
    if(!paperTile_datau)
        return;

    // Aligned data pointer.
    BYTE * paperTile_data = paperTile_datau + (32 - (DWORD)paperTile_datau % 32);

    // Fill the image data.
    const float * src = (const float*) color->lock_pointer_const ();
    for (int y = 0, i = 0; y < color->height (); y++) {
        BYTE * line = paperTile_data + y * 4 * color->width ();
        for (int x = 0; x < color->width (); x++, i++) {
            BYTE texel = (BYTE)(src[i] * 255.0f);
            line[x * 3 + 0] = texel;
            line[x * 3 + 1] = texel;
            line[x * 3 + 2] = texel;
            line[x * 3 + 3] = 255;
        }
    }

    color->unlock ();

    // Cast to QImage.
    QImage * paperTile = new QImage (
        paperTile_data,
        color->width (),
        color->height (),
        color->width () * 4,
        QImage::Format_RGB888
    );

    // Add to combo at last.
    if (paperTile) {
        if (combo) {
            combo->addItem (QIcon (QPixmap::fromImage (*paperTile)), dscr->name);
        }
        delete paperTile;
    }

    SAFE_DELETE_A (paperTile_datau);
}

/// Save the interface pointer and
/// request the list of available
/// paper textures.
void PaperIndicator::connect (IConnector * caller) {
    SAFE_RELEASE (m_Connector);

    if (caller) {
        clear ();

        m_Connector = caller;
        m_Connector->addRef ();
        m_Connector->allocateObject (this, 0, ENTITY_PAPER_TEXTURE);
    }
}

void PaperIndicator::complitedRequest (int param, int length, void * data) {
    Q_UNUSED (length);

    if (param == 0) {
        sPaper * paper = static_cast<sPaper*>(data);
        if (paper) {
            m_TexturesPack.append (paper);
            setPaper (paper);

            emit paperChanged (paper);
        }
    }
}

void PaperIndicator::comboChanged (int i) {
    if (i >= 0 && i < m_TexturesPack.size ())
        emit paperChanged (m_TexturesPack[i]);
}

void PaperIndicator::setByName (QString name) {
    if (!combo)
        return;

    // Find a paper by name,
    for (int i = 0; i < m_TexturesPack.size (); i++) {
        sPaper * paper = m_TexturesPack[i];

        if (paper->name == name) {
            combo->setCurrentIndex (i);
            return;
        }
    }
}

void PaperIndicator::setByClass (sPaper::ePaperClass c) {
    if (!combo)
        return;

    combo->clear ();

    // Resort.
    for (int i = 0;;) {
        for (int j = 0; j < m_TexturesPack.size (); j++) {
            if (m_TexturesPack[j]->paper_class == c) {
                m_TexturesPack.swap (i, j);
                i++;
            }
        }

        break;
    }

    // Refill the combo.
    foreach (sPaper * paper, m_TexturesPack) {
        setPaper (paper);
    }

    // Find a paper by class,
    for (int i = 0; i < m_TexturesPack.size (); i++) {
        sPaper * paper = m_TexturesPack[i];

        if (paper->paper_class == c) {
            combo->setCurrentIndex (i);
            return;
        }
    }
}

unsigned long PaperIndicator::requiredBufferSize (void) const {
    const int i = combo->currentIndex ();
    if (i >= 0) {
        return m_TexturesPack[i]->name.toUtf8 ().size ();
    }

    return 0L;
}

unsigned long PaperIndicator::saveData (QDataStream * stream, unsigned long capacity) {
    unsigned long sz = 0L;
    if (capacity >= requiredBufferSize ()) {
        const int i = combo ? combo->currentIndex () : 0;
        if (i >= 0) {
            QByteArray data = m_TexturesPack[i]->name.toUtf8 ();

            const unsigned long toCopy = data.size ();
            const char * p = data.data ();

            stream->operator << ((qint32)toCopy);
            for (unsigned long r = 0; r < toCopy; r += stream->writeRawData (p + r, toCopy - r)) ;
            sz += toCopy + sizeof(qint32);
        }
    }

    return sz;
}

unsigned long PaperIndicator::loadData (QDataStream * stream, unsigned long capacity) {
    qint32 toCopy = 0;
    stream->operator >> (toCopy);

    QByteArray data (toCopy + 1, 0);
    char * p = data.data ();
    for (int r = 0; r < toCopy; r += stream->readRawData (p + r, toCopy - r)) ;

    // Set the paper by the name we (I hope) have loaded.
    setByName (QString::fromUtf8 (data.data (), toCopy));

    return toCopy + sizeof(qint32);
}
