/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef INSTRUMENTPROXY_H
#define INSTRUMENTPROXY_H

#include "commtypes.h"
#include "commdefs.h"
#include "commmath.h"
#include "refcounter.h"

//#ifdef QT_OPENGL_LIB
#   include "../../glew/include/GL/glew.h"
#   include "iInstrument.h"
#   include "iInstrumentProxy.h"

#include <time.h>
#include <QStack>

#define BRUSH_QUEUE_LENGTH          3
#define BRUSH_PRESS_QUEUE_LENGTH    16

/**
 * Define this to disable the interpolation
 * between the stroke points.
 *
 * #define DEBUG_DISABLE_STROKE_INTERPOLATION
 */


/**
 * Instrument proxy state block.
 */
struct sState {
    float pressure;                       /**< Last instrument pressure.      */
    int   xQueue[BRUSH_QUEUE_LENGTH + 1]; /**< Last instrument x-positions.   */
    int   yQueue[BRUSH_QUEUE_LENGTH + 1]; /**< Last instrument y-positions.   */
    int   nQueued;

    long  meter;                          /**< Travelled distance.                */
    long  anchor_x;                       /**< Brush anchor for a distance meter. */
    long  anchor_y;                       /**< Brush anchor for a distance meter. */

    float pressQueue[BRUSH_PRESS_QUEUE_LENGTH];
    int   nQueued_p;

    float  velocity[2];
};

/**
 * A proxy class for an
 * instrumnt plugin.
 */
class InstrumentProxy : public QObject,
                        public IInstrumentProxy,
                        public CRefCounter {
    Q_OBJECT

protected:
    IInstrument * m_Instrument;

    clock_t m_Timeshot;

    int    m_Id;
    float  m_Jitter;
    float  m_ScaleModifier;
    float  m_SensitivityModifier;

    /**
     * Pressure mapping curve.
     */
    ISpline * m_PressureCurve;

    /**
     * Proxy's state stack.
     */
    QStack<sState> m_States;

    InstrumentProxy (IInstrument * instrument);
    virtual ~InstrumentProxy (void);

    virtual void suicide (void) {
        delete this;
    }

friend IInstrumentProxy * g_fac_SPAWN_INSTRUMENT_PROXY (IInstrument * instrument);
public:

    /******* IINSTRUMENTPROXY IMPLANTS *********/

    /**
     * Get the scaling factor modifier
     * associated with the instrument.
     */
    virtual float scaleModifier (void) const {
        return m_ScaleModifier;
    }

    /**
     * Get the texture sensitivity modifier
     * associated with the instrument.
     */
    virtual float textureSensitivityModifier (void) const {
        return m_SensitivityModifier;
    }

    /**
     * Set the scaling factor modifier
     * associated with the instrument.
     */
    virtual void setScaleModifier (float scale) {
        m_ScaleModifier = scale;
    }

    /**
     * Set the texture sensitivity modifier
     * associated with the instrument.
     */
    virtual void setTextureSensitivityModifier (float sensitivity) {
        m_SensitivityModifier = MIN (0.95f, sensitivity);
    }

    /**
     * Get the X-component of
     * current brush velocity.
     */
    virtual float velocityX (void) const {
        if (!m_States.isEmpty ())
             return m_States.top ().velocity[0];
        else
            return 0.0f;
    }

    /**
     * Get the Y-component of
     * current brush velocity.
     */
    virtual float velocityY (void) const {
        if (!m_States.isEmpty ())
             return m_States.top ().velocity[1];
        else
            return 0.0f;
    }

    /**
     * Set the amount of jitter to
     * apply to an instrument.
     */
    virtual void applyJitter (float amount) {
        m_Jitter = amount;
    }

    /**
     * Notify a proxy when the instrument
     * is attached to current pipeline.
     */
    virtual void attached (void);

    /**
     * Notify a proxy when the instrument
     * is detached from current pipeline.
     */
    virtual void detached (void);

    /**
     * Get the current jitter value.
     */
    virtual float jitter (void) const {
        return m_Jitter;
    }

    virtual const void * qimage (void) {
        if (m_Instrument)
            return m_Instrument->image ();
        else
            return NULL;
    }

    /**
     * Reset the brush traveling meter.
     */
    virtual void meterReset (void) {
        if (!m_States.isEmpty ())
             m_States.top ().meter = 0L;
    }

    /**
     * Reset the brush anchored meter.
     */
    virtual void anchoredMeterReset (void);

    /**
     * Brush traveled distance in pixels.
     */
    virtual long meter (void) const {
        if (m_States.isEmpty ())
            return 0L;
        else
            return m_States.top ().meter;
    }

    /**
     * Brush distance from the last anchor in pixels.
     */
    virtual long anchoredMeter (void) const;

    virtual ISpline * pressureCurve (void) {
        return m_PressureCurve;
    }

    /**
     * Get the underlying instrument object.
     */
    virtual IInstrument * soul (void) {
        return m_Instrument;
    }

    /*********** IINSTRUMENT IMPLANTS *************/

    virtual const void * image (void) {
        if (m_Instrument)
            return m_Instrument->image ();
        else
            return NULL;
    }

    /**
     * Apply the instrument.
     * The delegate class provides the methods
     * for interactng with a canvas.
     */
    virtual void applyInstrument (IChannel * channel, ICanvasDelegate * delegate) {
        if (m_Instrument)
            return m_Instrument->applyInstrument (channel, delegate);
    }

    /**
     * Generate a copy of this instrument.
     */
    virtual IInstrument * replicate (void) {
        // FIXME Replicate the proxy state as well.
        if (m_Instrument)
            return m_Instrument->replicate ();
        else
            return NULL;
    }

    /**
     * Change theinstrument position.
     * The coordinates are given in pixels
     * is canvas space.
     */
    virtual void setPosition (int x, int y, eInputMethod source);

    /**
     * Get the point of last stoke.
     * @param alpha Weight [0, 1].
     * @param x Out: x-position between last two points.
     * @param y Out: y-position between last two points.
     */
    virtual void stroke (float alpha, int & x, int & y);

    /**
     * Reset all accumulated information.
     * This method is usually called on the
     * end of a stroke.
     */
    virtual void reset (void);

    /**
     * Change the brush pressure.
     */
    virtual void setPressure (float pressure, eInputMethod source);

    /**
     * Get current pressure value.
     */
    virtual float pressure (void) {
        if (m_States.isEmpty ())
            return 0.0f;
        else
            return m_States.top ().pressure;
    }

    /**
     * Change the brush tilt.
     */
    virtual void setTilt (float xtilt, float ytilt, eInputMethod source) {
        if (m_Instrument)
            m_Instrument->setTilt (xtilt, ytilt, source);
    }

    /**
     * Get the width of current
     * writing primitive.
     */
    virtual int currentPrimitiveWidth (void) {
        if (m_Instrument)
            return (int)(scaleModifier () * m_Instrument->currentPrimitiveWidth ());
        else
            return 0;
    }

    /**
     * Get the height of current
     * writing primitive.
     */
    virtual int currentPrimitiveHeight (void) {
        if (m_Instrument)
            return (int)(scaleModifier () * m_Instrument->currentPrimitiveHeight ());
        else
            return 0;
    }

    /**
     * Absorbing component of current
     * color mixture or its average if the
     * single value is not available.s
     */
    virtual const float * currentAbsorbing (void) {
        if (m_Instrument)
            return m_Instrument->currentAbsorbing ();
        else
            return NULL;
    }

    /**
     * Scattering component of current
     * color mixture or its average if the
     * single value is not available.s
     */
    virtual const float * currentScattering (void) {
        if (m_Instrument)
            return m_Instrument->currentScattering ();
        else
            return NULL;
    }

    /**
     * Returns the group this instrument
     * belongs to.
     */
    virtual eInstrumentGroup instrumentGroup (void) {
        if (m_Instrument)
            return m_Instrument->instrumentGroup ();
        else
            return INSTRUMENTGROUP_UNGROUPED;
    }

    /**
     * Get the promoted channels by index.
     * Promoted channels are used to request the canvas
     * to ensure that it contains all nessesary channels
     * when the instrument is to be applied.
     * The plugin should use it to promote all channels
     * whose presence it considers important.
     * /return Channel uid or NULL if the index is out of range
     */
    virtual FOURCC enumPromotedChannels (int index) {
        if (m_Instrument)
            return m_Instrument->enumPromotedChannels (index);
        else
            return FOURCC_NULL;
    }

    /**
     * Get the preferred color chooser for
     * this instrument.
     */
    virtual eColorMetod promoteColorChooser (void) const {
        if (m_Instrument)
            return m_Instrument->promoteColorChooser ();
        else
            return COLORMETHOD_KMSAMPLE;
    }

    /**
     * Add the given pigment mixture
     * to the instrument.
     */
    virtual void addMass (IPigmentMixture * pigment, float amount);

    /**
     * Removes all components
     * from the instrument.
     */
    virtual void cleanMass (void) {
        if (m_Instrument)
            m_Instrument->cleanMass ();
    }

    /**
     * Save the current brush geometrical
     * state in a stack.
     */
    virtual void statePush (void);

    /**
     * Restore brush state from a stack.
     */
    virtual void statePop (void);

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const {
        if (m_Instrument)
            return m_Instrument->changesCheck ();
        else
            return FALSE;
    }

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void) {
        if (m_Instrument)
            m_Instrument->changesCommit ();
    }

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const {
        if (m_Instrument)
            return m_Instrument->requiredBufferSize ();
        else
            return 0;
    }

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity) {
        if (m_Instrument)
            return m_Instrument->saveData (stream, capacity);
        else
            return 0L;
    }

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity) {
        if (m_Instrument)
            return m_Instrument->loadData (stream, capacity);
        else
            return 0L;
    }

    /************ IPLUGIN IMPLANTS ***************/

    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Get the name of this plugin.
     */
    virtual const char * name (void) {
        if (m_Instrument)
            return m_Instrument->name ();
        else
            return NULL;
    }

    /**
     * Get the description of this plugin.
     */
    virtual const char * description (void) {
        if (m_Instrument)
            return m_Instrument->description ();
        else
            return NULL;
    }

    /**
     * Get the author of this plugin.
     */
    virtual const char * author (void) {
        if (m_Instrument)
            return m_Instrument->author ();
        else
            return NULL;
    }

    /**
     * Get the number of custom parameters
     * accepted by the plugin.
     */
    virtual int customParametersCount (void) const {
        if (m_Instrument)
            return m_Instrument->customParametersCount ();
        else
            return 0;
    }

    /**
     * Get the name of custom parameter
     * with the given index.
     * /return Pointer to string of NULL if the index is out of range.
     */
    virtual const char * customParameterName (int index) const {
        if (m_Instrument)
            return m_Instrument->customParameterName (index);
        else
            return NULL;
    }

    /**
     * Get the type of custom parameter
     * with the given index.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (int index) const {
        if (m_Instrument)
            return m_Instrument->customParameterType (index);
        else
            return PARAM_UNDEFINED;
    }

    /**
     * Get the type of custom parameter
     * with the given name.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (const char * name) const {
        if (m_Instrument)
            return m_Instrument->customParameterType (name);
        else
            return PARAM_UNDEFINED;
    }

    /**
     * Get the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully retrieved.
     */
    virtual bool customParameterGet (const char * name, void * buffer) {
        if (m_Instrument)
            return m_Instrument->customParameterGet (name, buffer);
        else
            return FALSE;
    }

    /**
     * Set the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully set.
     */
    virtual bool customParameterSet (const char * name, const void * buffer) {
        if (m_Instrument)
            return m_Instrument->customParameterSet (name, buffer);
        else
            return FALSE;
    }

    /**
     * Deinit all OpenGL-related data
     * structures to be prepared for a
     * context switching.
     */
    virtual void deInit (void) {
        if (m_Instrument)
            m_Instrument->deInit ();
    }

    /******* ISPAWNEDOBJECT IMPLANTS ********/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        if (m_Instrument)
            return m_Instrument->myClassName ();
        else
            return NULL;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

signals:
    /**
     * This signal is emitted when instrument
     * wants to change the color of color-chooser
     * to its own value.
     */
    void colorPromoted (const float * absorbing, const float * scattering);
};

#endif // INSTRUMENTPROXY_H
