/***
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "work_canvasiterate.h"


Work_CanvasIterate::Work_CanvasIterate (ICanvas       * canvas,
                                        IOpenGLHelper * helper,
                                        int  priority,
                                        bool persistent,
                                        sWorkerParameters * params) {
    m_Canvas     = canvas;
    m_Helper     = helper;
    m_Priority   = priority;
    m_Persistent = persistent;
    m_Connector  = NULL;
    m_Logger     = NULL;
    m_Id         = ISpawnedObject::allocateId ();
    m_Active     = TRUE;
    m_Params     = params;
    m_LastError  = NULL;

    m_LoggingEnabled = FALSE;

    if (canvas) canvas->addRef ();
    if (helper) helper->addRef ();

    m_Time.restart ();
}

Work_CanvasIterate::~Work_CanvasIterate (void) {
    foreach (IPipeline * pipe, m_Pipelines) {
        if (pipe)
            pipe->release ();
    }

    logMessage (LOGIMPORTANCE_NEVERMIND, "Work complited -- canvas iterate.");

    SAFE_RELEASE (m_Canvas);
    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Connector);
    SAFE_RELEASE (m_Logger);
}

void Work_CanvasIterate::connect (IConnector * caller) {
    SAFE_RELEASE (m_Connector);
    if (caller) {
        m_Connector = caller;
        m_Connector->addRef ();

        foreach (IPipeline * pipe, m_Pipelines) {
            if (pipe)
                pipe->release ();
        }

        // Request the pipelines list.
        m_Pipelines.clear ();
        m_Connector->allocateObject (this, 13, ENTITY_PIPELINE_PLUGIN);
    }
}

void Work_CanvasIterate::complitedRequest (int param, int length, void * data) {
    Q_UNUSED (length);

    if (param == 13) {
        IPipeline * pipe = static_cast<IPipeline*>(data);
        if (pipe) {
            m_Pipelines.append (pipe);
            logMessage (LOGIMPORTANCE_NEVERMIND, "Attached pipeline: %s.", pipe->myClassName ());
        }
    }
}

bool Work_CanvasIterate::execute (void) {
    m_LastError = NULL;

    if (m_Active == FALSE) {
        return FALSE;
    }

    bool updated = FALSE;
    const float dt = MIN (0.1f, 0.001f * m_Time.restart ());
    if (m_Canvas) {
        if (m_Params->scanIdle) {
            // Precess the patches if we're not
            // in a scabbing mode.

            // For each pipeline.
            foreach (IPipeline * pipe, m_Pipelines) {
                if (pipe) {
                    // Try to apply the parameters by
                    // feeding them all at once.
                    if (m_Params) {
                        if (m_Params->paint) {
                            const float speed     = m_Params->paint->paintDryingSpeed ();
                            const float thickness = m_Params->paint->paintViscosity ();
                            pipe->customParameterSet ("Drying speed",    &speed);
                            pipe->customParameterSet ("Paint thickness", &thickness);
                        }

                        pipe->customParameterSet ("Flow speed",      &m_Params->flow_speed);
                        pipe->customParameterSet ("Granulation",     &m_Params->granulation);
                        pipe->customParameterSet ("Gravity",          m_Params->gravity);
                        pipe->customParameterSet ("Paper",           &m_Params->paper->height);
                        pipe->customParameterSet ("Paper offset",     m_Params->paper_offset);
                        pipe->customParameterSet ("Paper scale",      m_Params->paper_scale);
                    }

                    // And call the pipeline.
                    updated |= m_Canvas->applyPipeline (pipe, dt);
                }
            }
        } else {
            while (TRUE) {
                // Lazily analyze the patches one by one.
                IArrayGridPatch * patch = m_Canvas->patchIterator (m_Params->scanPatch);
                if (!patch) {
                    // All patches have been analyzed.
                    m_Params->scanIdle = TRUE;
                    break;
                } else {
                    patch->checkStatus ();
                    m_Params->scanPatch++;
                    m_Params->scanIdle = FALSE;
                    break;
                }
            }
        }
    } else {
        m_LastError = "No canvas is attached, staying idle.";
    }
    
    return updated;
}
