/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "worker.h"
#include <QtAlgorithms>


static bool workLessThan (const sTaskPack * w1, const sTaskPack * w2) {
    return w1->work->priority () > w2->work->priority ();
}

Worker::Worker (void) {
    m_IsIdle = TRUE;

    connect (&m_Timer, SIGNAL(timeout()), this, SLOT(itistime()));

    m_Params.granulation     = 0.0f;
    m_Params.gravity[0]      = 0.0f;
    m_Params.gravity[1]      = 0.0f;
    m_Params.paper_offset[0] = 0.0f;
    m_Params.paper_offset[1] = 0.0f;
    m_Params.paper_scale[0]  = 1.0f;
    m_Params.paper_scale[1]  = 1.0f;
    m_Params.paper           = NULL;
    m_Params.canvas          = NULL;
    m_Params.paint           = NULL;
    m_Params.scanPatch       = 0;
    m_Params.scanIdle        = TRUE;
    m_Params.flow_speed      = 1.0f;

    m_GlobalStatus = QHASGlobal::STATUS_IDLE;

    m_ScanningFrom.restart ();
    m_IdleFrom.restart ();
}

Worker::~Worker (void) {
    foreach (sTaskPack * task, m_Tasks) {
        if (task) {
            task->work->release ();
            delete task;
        }
    }

    SAFE_RELEASE (m_Params.canvas);
}

void Worker::userStartedPainting (void) {
    m_IsIdle = FALSE;

    // Change the status.
    if (m_GlobalStatus != QHASGlobal::STATUS_PRIORITY_USER) {
        m_GlobalStatus  = QHASGlobal::STATUS_PRIORITY_USER;

        emit statusChanged (QHASGlobal::STATUS_PRIORITY_USER);
    }

    // Reset the scanning.
    m_Params.scanIdle  = FALSE;
    m_Params.scanPatch = 0;
}

void Worker::userStoppedPainting (void) {
    m_IsIdle = TRUE;
    m_IdleFrom.restart ();

    // Report the idle state if
    // the timer is stopped.
    if (!m_Timer.isActive ()) {
        if (m_GlobalStatus != QHASGlobal::STATUS_IDLE) {
            m_GlobalStatus  = QHASGlobal::STATUS_IDLE;

            emit statusChanged (QHASGlobal::STATUS_IDLE);
        }
    }
}

void Worker::paperChanged (sPaper * paper) {
    if (paper) {
        m_Params.paper = paper;

        if (m_Params.canvas) {
            m_Params.paper_scale[0] = (float) m_Params.canvas->realWidth  () / paper->height->width  ();
            m_Params.paper_scale[1] = (float) m_Params.canvas->realHeight () / paper->height->height ();
        }
    }
}

void Worker::itistime (void) {
	PROFILE_FUNC();
    if (m_IsIdle &&
        m_IdleFrom.elapsed () > WORKER_IDLE_WAIT) {

        QHASGlobal::eStatus new_status = QHASGlobal::STATUS_IDLE;

        // Execute all tasks.
        if (m_Tasks.size ()) {
            int active = 0;
            foreach (sTaskPack * task, m_Tasks) {
                if (task) {
                    if (task->work->isActive () &&
                        task->timer.elapsed () >= task->work->period ())
                        active++;
                }
            }

            if (active) {
                // Perform the tasks.
                active = 0;
                foreach (sTaskPack * task, m_Tasks) {
                    if (task->timer.elapsed () >= task->work->period ()) {
                        if (task->work->execute ())
                            active++;

                        task->timer.restart ();
                    }
                }

                if (active)
                    new_status = QHASGlobal::STATUS_PRIORITY_SYSTEM;
            }
        }

        // Poke the scanner if needed.
        if (m_ScanningFrom.elapsed () > WORKER_RESCAN_WAIT) {
            m_ScanningFrom.restart ();

            m_Params.scanPatch = 0;
            m_Params.scanIdle  = FALSE;
        }

        // Change the status.
        if (m_GlobalStatus != new_status) {
            m_GlobalStatus  = new_status;

            emit statusChanged (new_status);
        }
    }
}

void Worker::start (void) {
    m_Timer.setInterval (WORKER_INTERVAL);
    m_Timer.start ();
}

void Worker::terminate (void) {
    m_Timer.stop ();

    // Release all tasks.
    foreach (sTaskPack * task, m_Tasks) {
        if (task) {
            task->work->release ();
            delete task;
        }
    }

    // Clear the list.
    m_Tasks.clear ();
    SAFE_RELEASE (m_Params.canvas);
}

IWork * Worker::addLayerPreview (IConnector    * connector,
                                 ICanvas       * canvas,
                                 ICanvasLayer  * layer,
                                 IOpenGLHelper * helper) {

    IWork * work = new Work_LayerPreview (canvas, layer, helper, 0, TRUE, &m_Params);
    if (work) {
        work->connect  (connector);
        m_Tasks.append (new sTaskPack (work));
    }

    qSort (m_Tasks.begin (), m_Tasks.end (), workLessThan);
    return work;
}

IWork * Worker::addCanvasUpdate (IConnector    * connector,
                                 ICanvas       * canvas,
                                 IOpenGLHelper * helper) {

    // Save the reference to canvas.
    SAFE_RELEASE (m_Params.canvas);
    if (canvas) {
        m_Params.canvas = canvas;
        m_Params.canvas->addRef ();
        if (m_Params.paper) {
            m_Params.paper_scale[0] = (float) canvas->realWidth  () / m_Params.paper->height->width  ();
            m_Params.paper_scale[1] = (float) canvas->realHeight () / m_Params.paper->height->height ();
        }

        IWork * work = new Work_CanvasIterate (canvas, helper, 0, TRUE, &m_Params);
        if (work) {
            work->connect  (connector);
            m_Tasks.append (new sTaskPack (work));
        }

        qSort (m_Tasks.begin (), m_Tasks.end (), workLessThan);
        return work;
    }

    return NULL;
}
