/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef WORKSPACE_H
#define WORKSPACE_H

#include "commtypes.h"
#include "commdefs.h"
#include "refcounter.h"

//#ifdef QT_OPENGL_LIB
#   include "../../glew/include/GL/glew.h"

#include "iWorkspace.h"

class Workspace : public IWorkspace,
                  public IPaintingStyle,
                  public CRefCounter {

    _VERBOSEOBJECT

public:

    /**
     * Get current canvas.
     */
    virtual ICanvas * canvas (void) {
        return m_Canvas;
    }

    /**
     * Get the current painting style.
     */
    virtual IPaintingStyle * paintingStyle (void) {
        return this;
    }

    /**
     * Create the workspace data.
     * Must be called only if spawned without
     * specified dimentions.
     */
    virtual void create (long width, long height, int dpi);

    /**
     * Attach the paer switch widget to
     * enable the saving/loading its state along
     * with workspace data.
     */
    virtual void attachPaperSwitch (IPaperSwitch * p) {
        m_PaperMonster = p;
    }

    /************ IPAINTINGSTYLE IMPLANTS ***************/

    /**
     * Set the painting style parameters
     * from the given preset.
     */
    virtual void fromPreset (ePAINTINGSTYLE_PRESET preset);

    /**
     * Promoted paint drying speed.
     */
    virtual float paintDryingSpeed (void) const {
        return m_ps_DryingSpeed;
    }

    /**
     * Promoted paint viscosity.
     */
    virtual float paintViscosity (void) const {
        return m_ps_Viscosity;
    }

    /**
     * Promoted paint opacity.
     */
    virtual float paintOpacity (void) const {
        return m_ps_Opacity;
    }

    /**
     * Promoted paint staining power.
     */
    virtual float paintStainingPower (void) const {
        return m_ps_Staining;
    }

    virtual void setDryingSpeed(float f) {
        m_ps_DryingSpeed = f;
    }
    virtual void setViscosity(float f) {
        m_ps_Viscosity = f;
    }
    virtual void setOpacity(float f) {
        m_ps_Opacity = f;
    }
    virtual void setStainingPower(float f) {
        m_ps_Staining = f;
    }

    /**
     * Find the preset closes to current state.
     */
    virtual ePAINTINGSTYLE_PRESET closestPreset (void) const;

    /************ IDATACONTAINER IMPLANTS ***************/

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const {
        return m_WorkspaceChanged;
    }

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void) {
        m_WorkspaceChanged = FALSE;
    }

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const;

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity);

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity);

    /******** IVERBOSEOBJECT IMPLANTS *********/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_LastError;
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Enable of disable logging.
     */
    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

    /********** ISPAWNEDOBJECT IMPLANTS ************/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::Workspace";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

friend IWorkspace * g_fac_SPAWN_WORKSPACE (ICanvas * canvas, IOpenGLHelper * helper, long width, long height, int dpi);
protected:
    bool            m_WorkspaceChanged;
    bool            m_WorkspaceCreated;

    int             m_Id;
    int             m_IOVersionMajor;
    int             m_IOVersionMinor;
    long            m_Width;
    long            m_Height;
    int             m_Dpi;
    ICanvas       * m_Canvas;
    IPaperSwitch  * m_PaperMonster;
    IOpenGLHelper * m_Helper;
    QWidget       * m_Parent;

    float m_ps_DryingSpeed;         /**< Painting style: drying speed.      */
    float m_ps_Opacity;             /**< Painting style: paint opacity.     */
    float m_ps_Viscosity;           /**< Painting style: viscosity.         */
    float m_ps_Staining;            /**< Painting style: staining power.    */

    Workspace (ICanvas * canvas, IOpenGLHelper * helper, long width, long height, int dpi);
    virtual ~Workspace (void);

    virtual void suicide (void) {
        delete this;
    }
};

//#endif // QT_OPENGL_LIB
#endif // WORKSPACE_H
