/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#pragma once

#include "commtypes.h"
#include "commdefs.h"
#include "commmath.h"
#include "refcounter.h"

#include "../../../glew/include/GL/glew.h"
#include <QtPlugin>

#include "iInstrument.h"
#include "iOpenGLHelper.h"
#include "iLogger.h"
#include "iChannel.h"
#include "iArray.h"

#include <QImage>
#include <QStack>
#include "brushtuft.h"
#include "instrument_brushes_global.h"


#define BRUSH_INNER_SIZE           32
#define PRIMITIVE_COLOR_CELLS      (INSTRUMENT_COLORBUFFER_SIZE * INSTRUMENT_COLORBUFFER_SIZE)
#define SLOTS_NUMBER               6

#define VALID_NOTHING               0
#define VALID_VERTICES              1
#define VALID_COLORS                2
#define VALID_EVERYTHING            3

enum eType {
    TYPE_BRUSH_ROUND = 0,
    TYPE_BRUSH_FLAT  = 1,
    TYPE_PENCIL      = 2,
    TYPE_ERASER      = 3
};

extern const char * g_NAMES_br[INSTRUMENTS_IN_PACK +
                               INSTRUMENTS_IN_TOOLPACK];
extern const char * g_CLASSES_br[INSTRUMENTS_IN_PACK +
                                 INSTRUMENTS_IN_TOOLPACK];

/**
 * Instrument locking slot.
 */
struct BrushLock : public IInstrumentLock {
    void     * m_Parent;
    int        m_ChannelIndex;
    int        m_Angle;
    int        m_nVBOvertices;
    int        m_Affection;
    bool       m_Locked;
    bool       m_Valid;
    float      m_Sensitivity;
    float      m_ComponentMask[4];
    float      m_BleedingRate;
    float      m_SuckingRate;
    float      m_ScalingFactor;
    float      m_Reservoir;
    IArray   * m_PrimitiveMask;
    IArray   * m_PrimitiveColor;
    IChannel * m_Channel;
    GLuint     m_VBO;

    sNoiseParameters m_NoiseParams;

    /**
     * Drawing privitive color texture.
     */
    virtual IArray * primitiveColor (void) {
        return m_PrimitiveColor;
    }

    /**
     * Drawing privitive shape texture.
     */
    virtual IArray * primitiveShape (void) {
        return m_PrimitiveMask;
    }

    /**
     * Drawing privitive geometry
     * as a vertex buffer.
     */
    virtual GLuint primitiveGeometry (int & nVertices) const {
        nVertices = m_nVBOvertices;
        return m_VBO;
    }

    /**
     * Rotation angle of the primitive
     * around the Y axis.
     */
    virtual int primitiveRotation (void) const {
        return m_Angle;
    }

    /**
     * Returns the compoenets mask that contain
     * zeros for unused components and the weights for
     * the rest.
     */
    virtual const float * flowControlComponentMask (void) const {
        return m_ComponentMask;
    }

    /**
     * Brush bleeding rate.
     */
    virtual float bleedingRate (void) const {
        return m_BleedingRate;
    }

    /**
     * Brush sucking rate.
     */
    virtual float suckingRate (void) const {
        return m_SuckingRate;
    }

    /**
     * Instrument reservoir size determines the
     * threshold delaying the color feedback.
     */
    virtual float reservoirSize (void) const {
        return m_Reservoir;
    }

    /**
     * Get the texture sensitivity of
     * the brush in a range [0, 1].
     */
    virtual float stipplingSensitivity (void) const {
        return m_Sensitivity;
    }

    /**
     * Perlin noise parameters block.
     */
    virtual const sNoiseParameters * noiseParameters (void) const {
        return &m_NoiseParams;
    }

    /**
     * Get the scaling factor for
     * the instrument.
     */
    virtual float scalingFactor (void) const {
        return m_ScalingFactor;
    }

    /**
     * OR-ed set of flags specifying
     * what data planes should be affected by instrument.
     */
    virtual int affectsDataPlanes (void) const {
        return m_Affection;
    }

    /**
     * Delete this descriptor and unlock the brush.
     * This method must be called once you finished
     * working with descriptor.
     */
    virtual void release (void);

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }
};


/**
 * Instrument state.
 */
struct sState {
    float x;
    float y;
    float pressure;
    float water;
};

/**
 * Default brushes pack.
 */
class BrushInstrument : public IInstrument,
                        public CRefCounter {

    BrushInstrument (int type,
                     IInstrumentFactory * factory,
                     IOpenGLHelper      * helper,
                     ILogger            * logger);

    virtual ~BrushInstrument (void);

    virtual void suicide (void) {
        delete this;
    }

    int   m_Id;                 /**< Object id.                         */
    int   m_Type;               /**< Object sub-type.                   */
    float m_Pressure;           /**< Current pressure.                  */
    float m_UpdateRate_suck;    /**< The speed of brush color updating. */
    float m_UpdateRate_bleed;   /**< The speed of brush color updating. */
    float m_StrokeAngle;        /**< Angle for the stroke texture.      */
    float m_Hardness;           /**< Hardness attribute (pencil only).  */
    int   m_VBOValid;
    bool  m_GLinited;
    bool  m_Changed;
    bool  m_Stroke;
    bool  m_PositionCached;     /**< Indicates that we have saved the last position.    */
    int   m_LastX;
    int   m_LastY;

    ILogger            * m_Logger;
    IOpenGLHelper      * m_Helper;
    IInstrumentFactory * m_Factory;

    /**
     * Instrument image.
     */
    QImage m_Image;

    /**
     * A stack of saved instruemnt states.
     */
    QStack<sState> m_StatesBk;

    IArray * m_Mask;                /**< Brush mask.                */
    IArray * m_MaskCtl;             /**< Brush ctl channel mask.    */

    IArray * m_Color;

    float m_WaterStorage;
    float m_AbsorbingAvg[4]  __attribute__ ((aligned (16)));
    float m_ScatteringAvg[4] __attribute__ ((aligned (16)));

    BrushTuft   m_Tuft;               /**< Brush-tuft model.            */
    sTuftNode * m_Tip;                /**< Brush tip.                   */

    BrushLock m_LockSlots[SLOTS_NUMBER];

    /**
     * Vertex and color buffers.
     */
    GLuint m_VBuf[4];
    int    m_nVertices;

friend class Instrument_brushes;
protected:

    void invalidate (int state) {
        m_VBOValid &= ~state;
    }

    void validate (void);
    void checkInited (void);
    void buildTuftModel (void);
    void initVertexBuffer (void);

    /**
     * Create the brush texture for a flat type.
     */
    void initBrushTexture_flat (IArray * mask, float randomness = 0.5f, bool alpha = TRUE);

    /**
     * Create the brush texture for a round type.
     */
    void initBrushTexture_round (IArray * mask, float randomness = 0.5f, bool alpha = TRUE);

    /**
     * Create the brush texture for a round hairy type.
     */
    void initBrushTexture_hairy (IArray * mask, float wet);

    void copyBrushTexture (IArray * src, IArray * dst, bool alpha = TRUE);

    void addMass (const float * a, const float * s, float amount);

#ifdef GLEW_MX
    GLEWContext * glewGetContext (void) {
         return m_Helper->context ();
     }
#endif

public:

    void slotUnlocked (BrushLock * slot);

    /******** IINSTRUMENT IMPLANTS  ***********/

    /**
     * Apply the instrument.
     * The delegate class provides the methods
     * for interactng with a canvas.
     */
    virtual void applyInstrument (IChannel * channel, ICanvasDelegate * delegate);

    /**
     * Generate a copy of this instrument.
     */
    virtual IInstrument * replicate (void);

    /**
     * Change the instrument position.
     * The coordinates are given in pixels
     * is canvas space.
     */
    virtual void setPosition (int x, int y, eInputMethod source);

    /**
     * Change the brush pressure.
     */
    virtual void setPressure (float pressure, eInputMethod source);

    /**
     * Change the brush tilt.
     */
    virtual void setTilt (float xtilt, float ytilt, eInputMethod source);

    /**
     * Get the width of current
     * writing primitive.
     */
    virtual int currentPrimitiveWidth (void) {
        if (m_Mask)
            return m_Mask->width ();
        else
            return 0;
    }

    /**
     * Get the height of current
     * writing primitive.
     */
    virtual int currentPrimitiveHeight (void) {
        if (m_Mask)
            return m_Mask->height ();
        else
            return 0;
    }

    /**
     * Absorbing component of current
     * color mixture or its average if the
     * single value is not available.
     */
    virtual const float * currentAbsorbing (void) {
        if (m_Type != TYPE_ERASER)
            return m_AbsorbingAvg;
        else
            return NULL;
    }

    /**
     * Scattering component of current
     * color mixture or its average if the
     * single value is not available.
     */
    virtual const float * currentScattering (void) {
        if (m_Type != TYPE_ERASER)
            return m_ScatteringAvg;
        else
            return NULL;
    }

    /**
     * Returns a picture to
     * display in a list.
     */
    virtual const void * image (void) {
        return & m_Image;
    }

    /**
     * Returns the group this instrument
     * belongs to.
     */
    virtual eInstrumentGroup instrumentGroup (void){
        switch (m_Type) {
            case TYPE_BRUSH_FLAT:
            case TYPE_BRUSH_ROUND:
                return INSTRUMENTGROUP_BRUSH;
                break;
            case TYPE_ERASER:
                return INSTRUMENTGROUP_ERASER;
                break;
            case TYPE_PENCIL:
                return INSTRUMENTGROUP_PENCIL;
                break;
            default:
                return INSTRUMENTGROUP_UNGROUPED;
                break;
        }
    }

    /**
     * Get the promoted channels by index.
     * Promoted channels are used to request the canvas
     * to ensure that it contains all nessesary channels
     * when the instrument is to be applied.
     * The plugin should use it to promote all channels
     * whose presence it considers important.
     * /return Channel uid or NULL if the index is out of range
     */
    virtual FOURCC enumPromotedChannels (int index) {
        switch (index) {
            case 0:
                return CHANNELUID_Optical;
            case 1:
                return CHANNELUID_Meta0;
            default:
                return FOURCC_NULL;
        };
    }

    /**
     * Get the preferred color chooser for
     * this instrument.
     */
    virtual eColorMetod promoteColorChooser (void) const {
        return COLORMETHOD_PALETTE;
    }

    /**
     * Add the given pigment mixture
     * to the instrument.
     */
    virtual void addMass (IPigmentMixture * pigment, float amount);

    /**
     * Removes all components
     * from the instrument.
     */
    virtual void cleanMass (void);

    /**
     * Save the current brush geometrical
     * state in a stack.
     */
    virtual void statePush (void);

    /**
     * Restore brush state from a stack.
     */
    virtual void statePop (void);

    /******** IDATACONTAINER IMPLANTS *******/

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const {
        return m_Changed;
    }

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void) {
        m_Changed = FALSE;
    }

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const;

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity);

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity);

    /********** IPLUGIN IMPLANTS ************/

    /**
     * Get the name of this plugin.
     */
    virtual const char * name (void) {
        return g_NAMES_br[m_Type];
    }

    /**
     * Get the description of this plugin.
     */
    virtual const char * description (void) {
        static const char * g_DESCRIPTIONS [] = {
            "Small round brush from the Default Brushes pack.",
            "Large round brush from the Default Brushes pack.",
            "Pencil the Default Brushes pack.",
            "Eraser from the Default Brushes pack."
        };

        return g_DESCRIPTIONS[m_Type];
    }

    /**
     * Get the author of this plugin.
     */
    virtual const char * author (void) {
        return "shadow-of-kyle";
    }

    /**
     * Get the number of custom parameters
     * accepted by the plugin.
     */
    virtual int customParametersCount (void) const;

    /**
     * Get the name of custom parameter
     * with the given index.
     * /return Pointer to string of NULL if the index is out of range.
     */
    virtual const char * customParameterName (int index) const;

    /**
     * Get the type of custom parameter
     * with the given index.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (int index) const {
        Q_UNUSED (index);

        return PARAM_UNDEFINED;
    }

    /**
     * Get the type of custom parameter
     * with the given name.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (const char * name) const {
        Q_UNUSED (name);

        return PARAM_UNDEFINED;
    }

    /**
     * Get the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully retrieved.
     */
    virtual bool customParameterGet (const char * name, void * buffer);

    /**
     * Set the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully set.
     */
    virtual bool customParameterSet (const char * name, const void * buffer);

    /**
     * Deinit all OpenGL-related data
     * structures to be prepared for a
     * context switching.
     */
    virtual void deInit (void);

    /******* ISPAWNEDOBJECT IMPLANTS *******/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return g_CLASSES_br[m_Type];
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

};
