/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#pragma once

#include "commtypes.h"
#include "commdefs.h"
#include "commmath.h"
#include "refcounter.h"

#include "../../../glew/include/GL/glew.h"
#include <QtPlugin>
#include <QImage>

#include "iInstrument.h"
#include "instrument_brushes_global.h"
#include "iPigmentDescription.h"

enum eToolType {
    TYPE_EYEDROPPER  = 0,   /** Я пипетка, ы!!      */
    TYPE_SELECTION   = 1
};

extern const char * g_NAMES_br[INSTRUMENTS_IN_PACK +
                               INSTRUMENTS_IN_TOOLPACK];
extern const char * g_CLASSES_br[INSTRUMENTS_IN_PACK +
                                 INSTRUMENTS_IN_TOOLPACK];


/**
 * Default tools pack.
 */
class ToolInstrument : public IInstrument,
                       public CRefCounter {

    ToolInstrument (int type,
                    IInstrumentFactory * factory,
                    IOpenGLHelper      * helper,
                    ILogger            * logger);

    virtual ~ToolInstrument (void);

    virtual void suicide (void) {
        delete this;
    }

    int   m_Id;                 /**< Object id.                         */
    int   m_Type;               /**< Object sub-type.                   */
    float m_Pressure;           /**< Current pressure.                  */
    bool  m_Stroke;
    bool  m_PositionCached;     /**< Indicates that we have saved the last position.    */
    int   m_LastX;
    int   m_LastY;

    ILogger             * m_Logger;
    IOpenGLHelper       * m_Helper;
    IInstrumentFactory  * m_Factory;
    IPigmentDescription * m_HoverPigment;

    /**
     * Instrument image.
     */
    QImage m_Image;

    // Data block for selection instrument.
    QPoint m_Selection_pivot;       /**< Selection corner.  */
    QPoint m_Selection_hand;        /**< Selection corner.  */
    bool   m_Selected;              /**< Selection trigger. */

friend class Instrument_brushes;
public:

    /******** IINSTRUMENT IMPLANTS  ***********/

    /**
     * Apply the instrument.
     * The delegate class provides the methods
     * for interactng with a canvas.
     */
    virtual void applyInstrument (IChannel * channel, ICanvasDelegate * delegate);

    /**
     * Generate a copy of this instrument.
     */
    virtual IInstrument * replicate (void);

    /**
     * Change the instrument position.
     * The coordinates are given in pixels
     * is canvas space.
     */
    virtual void setPosition (int x, int y, eInputMethod source);

    /**
     * Change the brush pressure.
     */
    virtual void setPressure (float pressure, eInputMethod source);

    /**
     * Change the brush tilt.
     */
    virtual void setTilt (float xtilt, float ytilt, eInputMethod source) {
        Q_UNUSED (xtilt);
        Q_UNUSED (ytilt);
        Q_UNUSED (source);
    }

    /**
     * Get the width of current
     * writing primitive.
     */
    virtual int currentPrimitiveWidth (void) {
        return 4;
    }

    /**
     * Get the height of current
     * writing primitive.
     */
    virtual int currentPrimitiveHeight (void) {
        return 4;
    }

    /**
     * Absorbing component of current
     * color mixture or its average if the
     * single value is not available.
     */
    virtual const float * currentAbsorbing (void) {
        return NULL;
    }

    /**
     * Scattering component of current
     * color mixture or its average if the
     * single value is not available.
     */
    virtual const float * currentScattering (void) {
        return NULL;
    }

    /**
     * Returns a picture to
     * display in a list.
     */
    virtual const void * image (void) {
        return & m_Image;
    }

    /**
     * Returns the group this instrument
     * belongs to.
     */
    virtual eInstrumentGroup instrumentGroup (void){
        return INSTRUMENTGROUP_TOOLBOX;
    }

    /**
     * Get the promoted channels by index.
     * Promoted channels are used to request the canvas
     * to ensure that it contains all nessesary channels
     * when the instrument is to be applied.
     * The plugin should use it to promote all channels
     * whose presence it considers important.
     * /return Channel uid or NULL if the index is out of range
     */
    virtual FOURCC enumPromotedChannels (int index) {
        Q_UNUSED (index);

        return FOURCC_NULL;
    }

    /**
     * Get the preferred color chooser for
     * this instrument.
     */
    virtual eColorMetod promoteColorChooser (void) const {
        return COLORMETHOD_KMSAMPLE;
    }

    /**
     * Add the given pigment mixture
     * to the instrument.
     */
    virtual void addMass (IPigmentMixture * pigment, float amount) {
    }

    /**
     * Removes all components
     * from the instrument.
     */
    virtual void cleanMass (void) {
    }

    /**
     * Save the current brush geometrical
     * state in a stack.
     */
    virtual void statePush (void) {
    }

    /**
     * Restore brush state from a stack.
     */
    virtual void statePop (void) {
    }

    /******** IDATACONTAINER IMPLANTS *******/

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const {
        return FALSE;
    }

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void) {
    }

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const {
        return 0L;
    }

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity) {
        Q_UNUSED (stream);
        Q_UNUSED (capacity);

        return 0L;
    }

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity) {
        Q_UNUSED (stream);
        Q_UNUSED (capacity);

        return 0L;
    }

    /********** IPLUGIN IMPLANTS ************/

    /**
     * Get the name of this plugin.
     */
    virtual const char * name (void) {
        return g_NAMES_br[INSTRUMENTS_IN_PACK + m_Type];
    }

    /**
     * Get the description of this plugin.
     */
    virtual const char * description (void) {
        static const char * g_DESCRIPTIONS [] = {
            "Eyedropper from the Default Brushes pack.",
            "Selection tool from the Default Brushes pack."
        };

        return g_DESCRIPTIONS[m_Type];
    }

    /**
     * Get the author of this plugin.
     */
    virtual const char * author (void) {
        return "shadow-of-kyle";
    }

    /**
     * Get the number of custom parameters
     * accepted by the plugin.
     */
    virtual int customParametersCount (void) const {
        return 0;
    }

    /**
     * Get the name of custom parameter
     * with the given index.
     * /return Pointer to string of NULL if the index is out of range.
     */
    virtual const char * customParameterName (int index) const {
        return NULL;
    }

    /**
     * Get the type of custom parameter
     * with the given index.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (int index) const {
        Q_UNUSED (index);

        return PARAM_UNDEFINED;
    }

    /**
     * Get the type of custom parameter
     * with the given name.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (const char * name) const {
        Q_UNUSED (name);

        return PARAM_UNDEFINED;
    }

    /**
     * Get the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully retrieved.
     */
    virtual bool customParameterGet (const char * name, void * buffer) {
        Q_UNUSED (name);
        Q_UNUSED (buffer);

        return FALSE;
    }

    /**
     * Set the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully set.
     */
    virtual bool customParameterSet (const char * name, const void * buffer) {
        Q_UNUSED (name);
        Q_UNUSED (buffer);

        return FALSE;
    }

    /**
     * Deinit all OpenGL-related data
     * structures to be prepared for a
     * context switching.
     */
    virtual void deInit (void) {
    }

    /******* ISPAWNEDOBJECT IMPLANTS *******/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return g_CLASSES_br[INSTRUMENTS_IN_PACK + m_Type];
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

};
