/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "pipeline_stage0.h"
#include <QMutexLocker>


static const char * const g_err_UNKNOWN_CHANNEL     = "This channel type is not accepted.";
static const char * const g_err_UNKNOWN_PARAMETER   = "Parameter name is unknown.";
static const char * const g_PARAM_NAMES[] = {
    "Drying speed",
    "Granulation",
    "Paper",
    "Paper offset",
    "Paper scale",
    "Paint thickness",
    "Flow speed"
};
static const ePLUGIN_PARAM_TYPE g_PARAM_TYPES[] = {
    PARAM_FLOAT,
    PARAM_FLOAT,
    PARAM_IARRAY,
    PARAM_VEC2,
    PARAM_VEC2,
    PARAM_FLOAT,
    PARAM_FLOAT
};


Pipeline_stage0::Pipeline_stage0 (IOpenGLHelper * helper) {
    m_Id = ISpawnedObject::allocateId ();
    ISpawnedObject::_register (this);

    m_Helper         = helper;
    m_glslCopy       = NULL;
    m_glslDry0       = NULL;
    m_glslDry1       = NULL;
    m_glslAdvect     = NULL;
    m_glslSelfAdvect = NULL;
    m_Logger         = NULL;
    m_LastError      = NULL;
    m_chOptical      = NULL;
    m_chMeta         = NULL;
    m_Paper          = NULL;
    if (m_Helper) {
        m_Helper->addRef ();
    }

    m_LoggingEnabled = FALSE;
    m_GLinited       = FALSE;
    m_LastStatus     = FALSE;

    m_DryingSpeed = 1.0f;
    m_FlowSpeed   = 1.0f;
    m_Granulation = 0.5f;
    m_Step        = 0.0f;
    m_Embossing   = 0.0f;
}

Pipeline_stage0::~Pipeline_stage0 (void) {
    ISpawnedObject::_unregister (this);

    logMessage (LOGIMPORTANCE_NEVERMIND, "Shutting down the pipeline (stage0).");
    detach ();

    SAFE_RELEASE (m_glslCopy);
    SAFE_RELEASE (m_glslDry0);
    SAFE_RELEASE (m_glslDry1);
    SAFE_RELEASE (m_glslAdvect);
    SAFE_RELEASE (m_glslSelfAdvect);
    SAFE_RELEASE (m_Paper);
    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Logger);
}

int Pipeline_stage0::parameterDehash (const char * name) const {
    for (INT i = 0; i < customParametersCount (); i++) {
        if (0 == strcmp (g_PARAM_NAMES[i], name))
            return i;
    }

    return -1;
}

const char * Pipeline_stage0::customParameterName (int index) const {
    if (index >= 0 && index < customParametersCount ()) {
        return g_PARAM_NAMES[index];
    } else
        return NULL;
}

ePLUGIN_PARAM_TYPE Pipeline_stage0::customParameterType (int index) const {
    if (index >= 0 && index < customParametersCount ()) {
        return g_PARAM_TYPES[index];
    } else
        return PARAM_UNDEFINED;
}

ePLUGIN_PARAM_TYPE Pipeline_stage0::customParameterType (const char * name) const {
    return customParameterType (parameterDehash (name));
}

bool Pipeline_stage0::customParameterGet (const char * name, void * buffer) {
    m_LastError = NULL;
    switch (parameterDehash (name)) {
        case 0:
            *((float*)buffer) = m_DryingSpeed;
            return TRUE;
        case 1:
            *((float*)buffer) = m_Granulation;
            return TRUE;
        case 2:
            m_Paper->addRef ();
            *((void**)buffer) = m_Paper;
            return TRUE;
        case 3:
            ((float*)buffer)[0] = m_PaperOffset[0];
            ((float*)buffer)[1] = m_PaperOffset[1];
            return TRUE;
        case 4:
            ((float*)buffer)[0] = m_PaperScale[0];
            ((float*)buffer)[1] = m_PaperScale[1];
            return TRUE;
        case 5:
            *((float*)buffer) = m_Embossing;
            return TRUE;
        case 6:
            *((float*)buffer) = m_FlowSpeed;
            return TRUE;
        default:
            m_LastError = g_err_UNKNOWN_PARAMETER;
            return FALSE;
    }
}

bool Pipeline_stage0::customParameterSet (const char * name, const void * buffer) {
    m_LastError = NULL;
    switch (parameterDehash (name)) {
        case 0:
            m_DryingSpeed = *((float*)buffer);
            return TRUE;
        case 1:
            m_Granulation = *((float*)buffer);
            return TRUE;
        case 2:
            SAFE_RELEASE (m_Paper);
            m_Paper = *((IArray**)buffer);
            if (m_Paper)
                m_Paper->addRef ();
            return TRUE;
        case 3:
            m_PaperOffset[0] = ((float*)buffer)[0];
            m_PaperOffset[1] = ((float*)buffer)[1];
            return TRUE;
        case 4:
            m_PaperScale[0] = ((float*)buffer)[0];
            m_PaperScale[1] = ((float*)buffer)[1];
            return TRUE;
        case 5:
            m_Embossing = *((float*)buffer);
            return TRUE;
        case 6:
            m_FlowSpeed = *((float*)buffer);
            return TRUE;
        default:
            m_LastError = g_err_UNKNOWN_PARAMETER;
            return FALSE;
    }
}

void Pipeline_stage0::reloadShaders (void) {
   if (m_Helper) {
       IOpenGLShader * sh = m_Helper->loadShader ("copy.glsl");
       if (sh) {
           SAFE_RELEASE (m_glslCopy);
           m_glslCopy = sh;
       }
       sh = m_Helper->loadShader ("pipeline_dry.glsl");
       if (sh) {
           SAFE_RELEASE (m_glslDry0);
           m_glslDry0 = sh;
       }
       sh = m_Helper->loadShader ("pipeline_drybump.glsl");
       if (sh) {
           SAFE_RELEASE (m_glslDry1);
           m_glslDry1 = sh;
       }
       sh = m_Helper->loadShader ("pipeline_advect.glsl");
       if (sh) {
           SAFE_RELEASE (m_glslAdvect);
           m_glslAdvect = sh;
       }
       sh = m_Helper->loadShader ("pipeline_selfadvect.glsl");
       if (sh) {
           SAFE_RELEASE (m_glslSelfAdvect);
           m_glslSelfAdvect = sh;
       }

       if (m_glslCopy) {
           m_glPar0_src = m_glslCopy->getUniformLocation ("src");
       }
       if (m_glslDry0) {
           m_glPar1_optical   = m_glslDry0->getUniformLocation ("optical");
           m_glPar1_dt        = m_glslDry0->getUniformLocation ("dt");
           m_glPar1_gran      = m_glslDry0->getUniformLocation ("granulation");
           m_glPar1_ballance1 = m_glslDry0->getUniformLocation ("ballance1");
           m_glPar1_ballance2 = m_glslDry0->getUniformLocation ("ballance2");
       }
       if (m_glslDry1) {
           m_glPar2_optical_center = m_glslDry1->getUniformLocation ("optical_center");
           m_glPar2_optical_left   = m_glslDry1->getUniformLocation ("optical_left");
           m_glPar2_optical_top    = m_glslDry1->getUniformLocation ("optical_top");
           m_glPar2_optical_right  = m_glslDry1->getUniformLocation ("optical_right");
           m_glPar2_optical_bottom = m_glslDry1->getUniformLocation ("optical_bottom");
           m_glPar2_source         = m_glslDry1->getUniformLocation ("source");
           m_glPar2_dt             = m_glslDry1->getUniformLocation ("dt");
           m_glPar2_embossing      = m_glslDry1->getUniformLocation ("embossing");
       }
       if (m_glslAdvect) {
           m_glPar3_src   = m_glslAdvect->getUniformLocation ("src");
           m_glPar3_meta0 = m_glslAdvect->getUniformLocation ("meta0");
           m_glPar3_txx   = m_glslAdvect->getUniformLocation ("texstep_x");
           m_glPar3_txy   = m_glslAdvect->getUniformLocation ("texstep_y");
           m_glPar3_dt    = m_glslAdvect->getUniformLocation ("dt");
           m_glPar3_dtx   = m_glslAdvect->getUniformLocation ("texoffset");
       }
       if (m_glslSelfAdvect) {
           m_glPar4_meta0 = m_glslSelfAdvect->getUniformLocation ("meta0");
           m_glPar4_txx   = m_glslSelfAdvect->getUniformLocation ("texstep_x");
           m_glPar4_txy   = m_glslSelfAdvect->getUniformLocation ("texstep_y");
           m_glPar4_dt    = m_glslSelfAdvect->getUniformLocation ("dt");
       }
    } else {
        SAFE_RELEASE (m_glslCopy);
        SAFE_RELEASE (m_glslDry0);
        SAFE_RELEASE (m_glslDry1);
        SAFE_RELEASE (m_glslAdvect);
        SAFE_RELEASE (m_glslSelfAdvect);
    }
}

void Pipeline_stage0::deInit (void) {
    detach ();

    SAFE_RELEASE (m_glslCopy);
    SAFE_RELEASE (m_glslDry0);
    SAFE_RELEASE (m_glslDry1);
    SAFE_RELEASE (m_glslAdvect);
    SAFE_RELEASE (m_glslSelfAdvect);
    SAFE_RELEASE (m_Paper);
    m_GLinited = FALSE;
}

void Pipeline_stage0::detach (void) {
    SAFE_RELEASE (m_chMeta);
    SAFE_RELEASE (m_chOptical);
}

bool Pipeline_stage0::acceptsChannel (FOURCC uid) {
    m_LastError = NULL;

    if (uid == CHANNELUID_Meta0 ||
        uid == CHANNELUID_Optical)
        return TRUE;
    else {
        return FALSE;
    }
}

bool Pipeline_stage0::pushChannel (IChannel * channel) {
    m_LastError = NULL;

    if (!channel) {
        m_LastError = g_err_UNKNOWN_CHANNEL;
        return FALSE;
    }

    if (channel->uid () == CHANNELUID_Meta0) {
        SAFE_RELEASE (m_chMeta);
        m_chMeta = channel;
        m_chMeta->addRef ();
    } else
    if (channel->uid () == CHANNELUID_Optical) {
        SAFE_RELEASE (m_chOptical);
        m_chOptical = channel;
        m_chOptical->addRef ();
    } else {
        m_LastError = g_err_UNKNOWN_CHANNEL;
        return FALSE;
    }

    return TRUE;
}

bool Pipeline_stage0::iterate (float dt) {
    PROFILE_FUNC();

    static const float g_MIN_STEP = 0.05f;
    static const float g_MAX_STEP = 1.00f;

    QMutexLocker locker (m_Helper->mutex ());
    m_LastError = NULL;

    m_Step += 0.25f * dt * m_DryingSpeed;
    if (dt > 10.0f)
        m_Step = g_MAX_STEP;
    if (m_Step < g_MIN_STEP) {
        detach ();

        // Duplicate the last status to
        // show that we're active anyhow.
        return m_LastStatus;
    }

    // Clamp the step value.
    m_Step = MIN (m_Step, g_MAX_STEP);
    if (m_chOptical == NULL) {
        detach ();
        m_Step = 0.0f;
        return (m_LastStatus = FALSE);
    }

    // Init the OpenGL stuff
    // when we're ready to go.
    if (!m_GLinited) {
        reloadShaders ();
        m_GLinited = TRUE;
    }

    if (NULL == m_glslDry0 ||
        NULL == m_glslDry1) {
        m_Step = 0.0f;
        return (m_LastStatus = FALSE);
    }

    GLuint transport0 = m_Helper->sharedPatchLock ();

    // Setup the modelview matrix.
    glMatrixMode   (GL_MODELVIEW);
    glPushMatrix   ();
    glLoadIdentity ();
    glTranslatef   (0.0f,  0.0f, -1.0f);

    // Setup the projection matrix.
    glMatrixMode   (GL_PROJECTION);
    glPushMatrix   ();
    glLoadIdentity ();

    glOrtho    (-0.5f, +0.5f, -0.5f, +0.5f, +0.0f, +1.0f);
    glScalef   (1.0f, 1.0f, 1.0f);
    glViewport (0, 0,
                m_Helper->sharedPatchSize (),
                m_Helper->sharedPatchSize ()
    );

    // Setup the shaders.
    m_glslDry0->useProgram ();
    m_glslDry0->uniform1i  (m_glPar1_optical, 2);
    m_glslDry0->uniform1f  (m_glPar1_dt,      m_Step);
    m_glslDry0->uniform1f  (m_glPar1_gran,    m_Granulation);

    m_glslDry1->useProgram ();
    m_glslDry1->uniform1i  (m_glPar2_source,         2);
    m_glslDry1->uniform1i  (m_glPar2_optical_center, 3);
    m_glslDry1->uniform1i  (m_glPar2_optical_left,   4);
    m_glslDry1->uniform1i  (m_glPar2_optical_top,    5);
    m_glslDry1->uniform1i  (m_glPar2_optical_right,  6);
    m_glslDry1->uniform1i  (m_glPar2_optical_bottom, 7);
    m_glslDry1->uniform1f  (m_glPar2_dt,             m_Step);
    m_glslDry1->uniform1f  (m_glPar2_embossing,      m_Embossing);

    m_glslCopy->useProgram ();
    m_glslCopy->uniform1i  (m_glPar0_src, 1);

    m_glslAdvect->useProgram ();
    m_glslAdvect->uniform1i (m_glPar3_meta0, 0);
    m_glslAdvect->uniform1i (m_glPar3_src,   1);
    m_glslAdvect->uniform2f (m_glPar3_txx,   0.002f, 0.000f);
    m_glslAdvect->uniform2f (m_glPar3_txy,   0.000f, 0.002f);

    m_glslSelfAdvect->useProgram ();
    m_glslSelfAdvect->uniform1i (m_glPar4_meta0, 2);
    m_glslSelfAdvect->uniform1f (m_glPar4_dt,    m_Step);
    m_glslSelfAdvect->uniform2f (m_glPar4_txx,   0.002f, 0.000f);
    m_glslSelfAdvect->uniform2f (m_glPar4_txy,   0.000f, 0.002f);

    // Bind the framebuffer for patch rendering.
    IFBOWrapper * fbo = m_Helper->sharedFBO (OPENGLHELPER_FBO_PATCH);
    if (!fbo)
        return (m_LastStatus = FALSE);

    fbo->bind ();
    GLenum backup = fbo->attachTexture (transport0, 1);

    IArrayGrid * src = m_chOptical->lock ();
    IArrayGrid * met = m_chMeta->lock ();
    if (!src || !met)
        return (m_LastStatus = FALSE);

    const bool drying_enabled = IS_SIGNIFICANT (m_DryingSpeed);
    const bool flow_enabled   = IS_SIGNIFICANT (m_FlowSpeed);
    bool updated = FALSE;

    // Propagate the optical values.
    for (int y = 0; y < src->nPatchesY (); y++) {
        for (int x = 0; x < src->nPatchesX (); x++) {
            // Process only if the patch is wet
            // or has a wet neighbour.
            // FIXME
            IArrayGridPatch * met_patch_center = met->patch (x + 0, y + 0);
            IArrayGridPatch * wet_patch_center = src->patch (x + 0, y + 0);
            IArrayGridPatch * wet_patch_left   = src->patch (x - 1, y + 0);
            IArrayGridPatch * wet_patch_right  = src->patch (x + 1, y + 0);
            IArrayGridPatch * wet_patch_top    = src->patch (x + 0, y - 1);
            IArrayGridPatch * wet_patch_bottom = src->patch (x + 0, y + 1);
            if (!(wet_patch_center->status () & PATCHSTAT_WET))
                continue;

            GLuint tex0_c = met_patch_center->data ()->lock_gltex ();
            GLuint tex1_c = wet_patch_center->data ()->lock_gltex ();
            GLuint tex1_l = wet_patch_left->data ()->lock_gltex ();
            GLuint tex1_r = wet_patch_right->data ()->lock_gltex ();
            GLuint tex1_t = wet_patch_top->data ()->lock_gltex ();
            GLuint tex1_b = wet_patch_bottom->data ()->lock_gltex ();
            if (tex0_c != UINT_UNDEFINED &&
                tex1_c != UINT_UNDEFINED &&
                tex1_l != UINT_UNDEFINED &&
                tex1_t != UINT_UNDEFINED &&
                tex1_r != UINT_UNDEFINED &&
                tex1_b != UINT_UNDEFINED) {
                m_Helper->bindTexture_ex (tex1_c, 2, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_l, 3, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_t, 4, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_r, 5, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_b, 6, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex0_c, 0, GL_NEAREST, TRUE);

                glDrawBuffer (backup);

                m_glslDry0->useProgram ();
                m_glslDry0->uniform1f (m_glPar1_ballance1, 0.0f);
                m_glslDry0->uniform1f (m_glPar1_ballance2, 0.0f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X0Y0));
                m_glslDry0->uniform1f (m_glPar1_ballance1, 1.0f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X1Y0));
                m_glslDry0->uniform1f (m_glPar1_ballance1, 0.0f);
                m_glslDry0->uniform1f (m_glPar1_ballance2, 1.0f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X0Y1));
                m_glslDry0->uniform1f (m_glPar1_ballance1, 1.0f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X1Y1));

                glDrawBuffer (fbo->attachTexture (tex1_c));
                m_Helper->bindTexture_ex (transport0, 1);

                m_glslCopy->useProgram ();
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1));
/*
                // Advect the optical field.
                m_glslAdvect->useProgram ();
                m_glslAdvect->uniform1f (m_glPar3_dt,  m_Step);
                m_glslAdvect->uniform2f (m_glPar3_dtx, 0.0f, 0.0f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X0Y0));
                m_glslAdvect->uniform2f (m_glPar3_dtx, 0.5f, 0.0f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X1Y0));
                m_glslAdvect->uniform1f (m_glPar3_dt,  0.0f);
                m_glslAdvect->uniform2f (m_glPar3_dtx, 0.0f, 0.5f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X0Y1));
                m_glslAdvect->uniform2f (m_glPar3_dtx, 0.5f, 0.5f);
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X1Y1));
*/
                met_patch_center->data ()->unlock ();
                wet_patch_center->data ()->unlock ();
                wet_patch_left->data ()->unlock ();
                wet_patch_right->data ()->unlock ();
                wet_patch_top->data ()->unlock ();
                wet_patch_bottom->data ()->unlock ();
                updated = TRUE;
            }
        }
    }

    // Update the Meta0 channel.
    for (int y = 0; y < src->nPatchesY (); y++) {
        for (int x = 0; x < src->nPatchesX (); x++) {
            // Process only if the patch is wet
            // or has a wet neighbour.
            IArrayGridPatch * src_patch        = met->patch (x, y);
            IArrayGridPatch * wet_patch_center = src->patch (x + 0, y + 0);
            IArrayGridPatch * wet_patch_left   = src->patch (x - 1, y + 0);
            IArrayGridPatch * wet_patch_right  = src->patch (x + 1, y + 0);
            IArrayGridPatch * wet_patch_top    = src->patch (x + 0, y - 1);
            IArrayGridPatch * wet_patch_bottom = src->patch (x + 0, y + 1);
            if (!(wet_patch_center->status () & PATCHSTAT_WET))
                continue;

            GLuint tex0   = src_patch->data ()->lock_gltex ();
            GLuint tex1_c = wet_patch_center->data ()->lock_gltex ();
            GLuint tex1_l = wet_patch_left->data ()->lock_gltex ();
            GLuint tex1_r = wet_patch_right->data ()->lock_gltex ();
            GLuint tex1_t = wet_patch_top->data ()->lock_gltex ();
            GLuint tex1_b = wet_patch_bottom->data ()->lock_gltex ();
            if (tex0   != UINT_UNDEFINED &&
                tex1_c != UINT_UNDEFINED &&
                tex1_l != UINT_UNDEFINED &&
                tex1_t != UINT_UNDEFINED &&
                tex1_r != UINT_UNDEFINED &&
                tex1_b != UINT_UNDEFINED) {
                m_Helper->bindTexture_ex (tex0, 1);
                m_Helper->bindTexture_ex (tex1_c, 3, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_l, 4, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_t, 5, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_r, 6, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (tex1_b, 7, GL_NEAREST, TRUE);
                m_Helper->bindTexture_ex (transport0, 2);

                glDrawBuffer (backup);

                // Backup the patch.
                m_glslCopy->useProgram ();
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1));

                // Render the sub-chanels.
                glDrawBuffer (fbo->attachTexture (tex0));

                // Propagate the paints bump map.
                m_glslDry1->useProgram ();
                glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X1Y0));
                
                // Advect the velocities field.
                //m_glslSelfAdvect->useProgram ();
                //glCallList (m_Helper->sharedList (OPENGLHELPER_LIST_QUAD1_X0Y0));

                src_patch->data ()->unlock ();
                wet_patch_center->data ()->unlock ();
                wet_patch_left->data ()->unlock ();
                wet_patch_right->data ()->unlock ();
                wet_patch_top->data ()->unlock ();
                wet_patch_bottom->data ()->unlock ();
                updated = TRUE;
            }
        }
    }

    m_chOptical->unlock ();
    m_chMeta->unlock ();

    fbo->attachTexture (transport0, 0);
    m_Helper->sharedPatchUnlock (transport0);

    m_Step = 0.0f;

#ifdef DEBUG_GLFINISH
    glFinish ();
#endif

    // Restore.
    glMatrixMode (GL_PROJECTION);
    glPopMatrix ();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix ();

    detach ();
    return (m_LastStatus = updated);
}
